from .bytearray import ByteArray

import base64
import hashlib

salt = "".join(map(chr, [
	247,  26, 166, 222, 143,  23, 118, 168,
	  3, 157,  50, 184, 161,  86, 178, 169,
	 62, 221,  67, 157, 197, 221, 206,  86,
	211, 183, 164,   5,  74,  13,   8, 176
]))
packet_keys = []
identification_keys = []
msg_keys = []

def HashPassword(password):
	global salt
	salted_hash = hashlib.sha256(password.encode("ISO8859_1")).hexdigest() + salt
	
	return base64.b64encode(hashlib.sha256(salted_hash.encode("ISO8859_1")).digest())

def SetPacketKeys(keys):
	global packet_keys, identification_keys, msg_keys
	packet_keys, identification_keys, msg_keys = keys

def encode_chunks(v, n):
	global identification_keys
	
	DELTA = 0x9e3779b9
	def MX():
		return int(((z>>5)^(y<<2)) + ((y>>3)^(z<<4))^(sum^y) + (identification_keys[(p & 3)^e]^z))

	y = v[0]
	sum = 0
	if n > 1:
		z = v[n - 1]
		q = int(6 + 52 / n)
	while q > 0:
		q -= 1
		sum = (sum + DELTA) & 0xffffffff
		e =  ((sum >> 2) & 0xffffffff) & 3
		p = 0
		while p < n - 1:
			y = v[p + 1]
			z = v[p] = (v[p] + MX()) & 0xffffffff
			p += 1
		y = v[0]
		z = v[n - 1] = (v[n - 1] + MX()) & 0xffffffff
	return v

def BlockCipher(packet):
	stack = packet.stack
	if len(stack) == 0:
		raise Exception("Block cipher algorithm can't be applied to an empty ByteArray.")
	
	while len(stack) < 8:
		stack += b"\x00"
	
	packet = ByteArray(stack)
	chunks = []
	while len(packet.stack) > 0:
		chunks.append(packet.readLong())
	
	chunks = encode_chunks(chunks, len(chunks))
	
	packet.writeShort(len(chunks))
	for chunk in chunks:
		packet.writeLong(chunk)
	
	return packet

def XorCipher(packet, fingerprint):
	global msg_keys
	
	stack = packet.stack
	new_stack = []
	
	for index, byte in enumerate(stack):
		fingerprint += 1
		new_stack.append((byte ^ msg_keys[fingerprint % 20]) & 255)
	
	return ByteArray(bytes(new_stack))


class ByteArray(object):
	def __init__(self, stack = b""):
		self.stack = stack
	
	def writeByte(self, *byte):
		self.stack += bytes(map(lambda n: int(n % 256), [*byte]))
		return self
	
	def writeShort(self, short):
		short = int(short)
		return self.writeByte((short >> 8) & 255, short & 255)
	
	def writeInt(self, integer):
		integer = int(integer)
		return self.writeByte((integer >> 16) & 255, (integer >> 8) & 255, integer & 255)
	
	def writeLong(self, long):
		long = int(long)
		return self.writeByte((long >> 24) & 255, (long >> 16) & 255, (long >> 8) & 255, long & 255)
	
	def writeUTF(self, utf):
		self.writeShort(len(utf))
		
		if type(utf) == str:
			utf = utf.encode()
		
		self.stack += utf
		return self
	
	def writeBigUTF(self, big_utf):
		self.writeInt(len(big_utf))
		self.stack += utf.encode()
		return stack
	
	def writeBool(self, boolean):
		self.writeByte(boolean)
		return self
	
	def readByte(self, bytesQuantity = 1):
		byteStack = self.stack[:bytesQuantity]
		self.stack = self.stack[bytesQuantity:]
		
		byteStack += b"\x00" * (bytesQuantity - len(byteStack))
		
		return byteStack if len(byteStack) != 1 else byteStack[0]
	
	def readShort(self):
		shortStack = self.readByte(2)
		return (shortStack[0] << 8) + shortStack[1]
	
	def readInt(self):
		intStack = self.readByte(3)
		return (intStack[0] << 16) + (intStack[1] << 8) + intStack[2]
	
	def readLong(self):
		longStack = self.readByte(4)
		return (longStack[0] << 24) + (longStack[1] << 16) + (longStack[2] << 8) + longStack[3]
	
	def readUTF(self):
		byte = self.readByte(self.readShort())
		
		if type(byte) == int:
			return chr(byte)
		return byte.decode()
	
	def readBigUTF(self):
		byte = self.readByte(self.readInt())
		
		if type(byte) == int:
			return chr(byte)
		return byte.decode()
	
	def readBool(self):
		return self.readByte() == 1


from .bytearray import ByteArray

import sys
import socket
import traceback

class ConnectionHandler(object):
	def __init__(self, name, call_event):
		self.call_event = call_event
		self.socket = None
		self.ip = ""
		self.packetID = 0
		self.port = 0
		self.name = name
		self.open = False
	
	def connect(self, ip, port):
		try:
			self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		except:
			print("[{name}]: socket can't be created.".format(name=self.name))
			sys.exit(0)
		
		self.socket.setblocking(0)
		self.socket.settimeout(3)
		
		host = socket.gethostbyname(ip)
		if host == None:
			print("[{name}] host could not be resolved.".format(name=self.name))
			sys.exit(0)
		try:
			self.socket.connect((host, port))
		except socket.timeout:
			print("[{name}]: connection timed out.".format(name=self.name))
			sys.exit(0)
		
		self.socket.settimeout(None)
		
		self.ip = ip
		self.port = port
		self.open = True
		
		self.call_event("on_connection_made", self)
	
	def __receive(self, length):
		result = self.socket.recv(length)
		if length > 0 and len(result) == 0:
			self.open = False
			self.socket.close()
			raise Exception("The connection has been closed.")
		
		return result
	
	def receive(self):
		stack_length_size = self.__receive(1)[0]
		
		if stack_length_size > 0:
			stack_length_bytes = self.__receive(stack_length_size)
			stack_length = 0
			for index, byte in enumerate(stack_length_bytes):
				stack_length += byte << (8 * (stack_length_size - index - 1))
			
			return self.__receive(stack_length)
		return b""
	
	def send(self, identifiers, alpha_packet):
		beta_packet = None
		gamma_packet = None
		
		if not isinstance(identifiers, (list, tuple)):
			raise Exception("Unknown identifiers type? Must be a list or tuple object!")
		
		if isinstance(alpha_packet, ByteArray):
			beta_packet = ByteArray(bytes(identifiers) + alpha_packet.stack)
		
		elif isinstance(alpha_packet, bytes):
			beta_packet = ByteArray(bytes(identifiers) + alpha_packet)
		
		elif isinstance(alpha_packet, int):
			beta_packet = ByteArray().writeByte([*identifiers, alpha_packet])
		
		elif isinstance(alpha_packet, (list, tuple)):
			beta_packet = ByteArray().writeByte(1, 1).writeUTF(
				"\x01".join(
					map(
						str,
						[chr(identifiers[0]) + chr(identifiers[1])] + alpha_packet
					)
				)
			)
		
		else:
			raise Exception("Unknown packet type? Must be a ByteArray, bytes, int, list or tuple object! (Identifiers: " + repr(identifiers) + ")")
		
		stack_length = len(beta_packet.stack)
		if stack_length < 256:
			gamma_packet = ByteArray().writeByte(1, stack_length)
		elif stack_length < 65536:
			gamma_packet = ByteArray().writeByte(2).writeShort(stack_length)
		elif stack_length < 16777216:
			gamma_packet = ByteArray().writeByte(3).writeInt(stack_length)
		else:
			raise Exception("The packet is too big! (Identifiers: " + repr(identifiers) + ")")
		
		gamma_packet.writeByte(self.packetID)
		self.packetID = (self.packetID + 1) % 100
		
		gamma_packet.stack += beta_packet.stack
		
		self.call_event("on_socket_raw_send", self, identifiers, alpha_packet)
		try:
			self.socket.send(gamma_packet.stack)
		except:
			self.open = False
			self.call_event("on_connection_close", self)


from distutils.core import setup
setup(
	name = 'transfromage',
	packages = ['transfromage'],
	version = '0.1',
	license = 'MIT',
	description = 'Client library for the Transformice game',
	author = 'Tocutoeltuco - Turkitutu',
	author_email = 'tocutoeltuco@gmail.com',
	url = 'https://github.com/Tocutoeltuco/transfromage',
	download_url = 'https://github.com/Tocutoeltuco/transfromage/archive/v_01.tar.gz',
	keywords = ['TRANFORMICE', 'CLIENT', 'THREADING', 'ATELIER801', 'EVENT'],
	install_requires=[],
	classifiers=[
		'Development Status :: 5 - Production/Stable',
		'Intended Audience :: Developers',
		'Topic :: Software Development :: Build Tools',
		'License :: OSI Approved :: MIT License',
		'Programming Language :: Python :: 3.6'
	]
)


import threading

class new(object):
	class SmartThread(object):
		def __init__(self, max_threads):
			self.max_threads = max_threads
			self.tasksQueue = []
		
		def tasksQueueLoop(self, main_connection):
			while main_connection.open:
				if threading.active_count() < self.max_threads and self.tasksQueue != []:
					next_task = self.tasksQueue.pop(0)
					
					threading.Thread(*next_task[0], **next_task[1]).start()
		
		def runParallelTask(self, *args, **kwargs):
			if threading.active_count() < self.max_threads:
				threading.Thread(*args, **kwargs).start()
			
			else:
				self.tasksQueue.append([args, kwargs])


from .connection import ConnectionHandler
from .bytearray import ByteArray
from .utilities import new
from .cipher import HashPassword, BlockCipher, XorCipher, SetPacketKeys

from urllib.request import urlopen, Request
from json import loads

import traceback
import time

class ApiEndpointException(Exception):
	"""Main Exception for the api endpoint errors.
	Thrown when the API can't get the game encryption keys."""
	
	pass

class Client(object):
	"""This is the connection between the server connection and the events.
	It allows you to set events (ex: on_socket_raw_receive) and run your bot."""
	
	def __init__(self, max_threads = 10):
		# Initializacion values
		
		self.main = ConnectionHandler("main", self.call_event)
		self.bulle = None
		
		self.running_events = []
		self.game_packet_keys = []
		self.events = {}
		self.received_authkey = 0
		self.last_heartbeat = 0
		self.game_auth_key = 0
		self.game_version = 0
		self.game_connection_key = ""
		
		self.SmartThread = new.SmartThread(max_threads)
	
	def parse_packet(self, connection, packet):
		"""This is the parse packets function.
		This is called so the most technical stuff (such as parsing auth keys, switching the bulle connection)
		are working by default."""
		
		CCC = (packet.readByte(), packet.readByte())
		# Reads the packet identifiers
		
		if CCC == (26, 3): # Correct handshake identifiers
			online_players = packet.readLong()
			connection.packetID = packet.readByte() # Sets the packet id of the connection
			community = packet.readUTF()
			country = packet.readUTF()
			self.received_authkey = packet.readLong() # Receives an authentication key, this will be parsed
			# when the login function is called
			
			self.last_heartbeat = time.time()
			self.SmartThread.runParallelTask(target = self.heartbeat_loop) # Starts the heartbeat loop
			
			community = ByteArray().writeByte(0).writeByte(0)
			self.main.send([8, 2], community) # Sets the connection community to EN
			
			os_info = ByteArray().writeUTF("en").writeUTF("Linux")
			os_info.writeUTF("LNX 29,0,0,140").writeByte(0)
			self.main.send([28, 17], os_info) # Sends the OS Info
		
		elif CCC == (44, 1): # Switch bulle identifiers
			bulle_id = packet.readLong()
			bulle_ip = packet.readUTF()
			
			self.bulle = ConnectionHandler("bulle", self.call_event)
			self.bulle.connect(bulle_ip, self.main.port)
			
			self.bulle.send(CCC, ByteArray().writeLong(bulle_id))
		
		elif CCC == (44, 22): # PacketId offset identifiers
			connection.packetID = packet.readByte() # Sets the packet id of the connection
	
	def event(self, function):
		"""This is the decorator for setting an event.
		All what this does is to make a wrapper so when any event throws any error, this handles it.
		Also, this registers the event to the `events` dictionary"""
		
		def wrapper(*args, **kwargs):
			try:
				return function(*args, **kwargs)
			except KeyboardInterrupt:
				self.close_all()
			except Exception as exc:
				if function.__name__ == "on_error":
					self.close_all()
					raise exc
				
				else:
					if not self.call_event("on_error", function.__name__, exc):
						self.close_all()
						raise exc
		
		wrapper.__name__ = function.__name__
		self.events[function.__name__] = wrapper
		return wrapper
	
	def login(self, player, password, start_room = "1"):
		"""This is the login function.
		Hashes your password, parses the authentication key, encrypts the packet and sends it to the server."""
		
		packet = ByteArray().writeUTF(player).writeUTF(HashPassword(password))
		packet.writeUTF("app:/TransformiceAIR.swf/[[DYNAMIC]]/2/[[DYNAMIC]]/4").writeUTF(start_room)
		packet.writeLong(self.received_authkey ^ self.game_auth_key)
		
		self.main.send([26, 8], BlockCipher(packet).writeByte(0))
	
	def sendRoomMessage(self, message):
		"""This is the function that sends a message to the room.
		Encrypts the packet and sends it to the server."""
		
		self.bulle.send([6, 6], XorCipher(ByteArray().writeUTF(message), self.bulle.packetID))
	
	def main_receive_loop(self):
		"""Receive loop for the main connection."""
		
		while self.main.open: # While the main connection is open
			packet = self.main.receive()
			
			self.call_event("on_socket_raw_receive", self.main, ByteArray(packet))
			self.parse_packet(self.main, ByteArray(packet))
	
	def bulle_receive_loop(self):
		"""Receive loop for the bulle conenction."""
		
		while self.main.open: # While the main connection is open
			if self.bulle != None and self.bulle.open: # If the bulle connection is open
				packet = self.bulle.receive()
				
				self.call_event("on_socket_raw_receive", self.bulle, ByteArray(packet))
				self.parse_packet(self.bulle, ByteArray(packet))
	
	def heartbeat_loop(self):
		"""Heart beat/Keep alive loop for the conenction."""
		
		while self.main.open: # While the main connection is open
			if (time.time() - self.last_heartbeat) >= 10: # If we are ten seconds or more later of the last heartbeat sent
				self.main.send([26, 26], ByteArray()) # Send another heartbeat
				
				if self.bulle != None and self.bulle.open:
					self.bulle.send([26, 26], ByteArray())
				
				self.last_heartbeat = time.time() # Set the last heartbeat sent to this time
				time.sleep(10)
	
	def close_all(self):
		"""Close function."""
		
		if self.bulle != None and self.bulle.open:
			self.bulle.open = False
			try:
				self.bulle.socket.close()
				self.call_event("on_connection_close", self.bulle)
			except:
				pass
		
		self.main.open = False
		try:
			self.main.socket.close()
			self.call_event("on_connection_close", self.main)
		except:
			pass
	
	def call_event(self, event_name, *args, **kwargs):
		"""This function calls an event.
		Returns True if the event was CALLED (this doesn't mean that the event finished the call)."""
		
		if event_name in self.events:
			self.SmartThread.runParallelTask(target = self.events[event_name], args = args, kwargs = kwargs)
			return True
		return False
	
	def loop_run(self):
		"""Main loop."""
		
		self.SmartThread.runParallelTask(target = self.main_receive_loop)
		self.SmartThread.runParallelTask(target = self.bulle_receive_loop)
		
		while self.main.open:
			pass
		
		self.close_all() # Closes all
	
	def get_keys(self, api_tfmid, api_token):
		"""Get keys function.
		This gets the keys of the tocu.tk API for the transformice connection."""
		
		result = loads(
			urlopen(
				Request(
					"https://api.tocu.tk/get_transformice_keys.php?tfmid=" + str(api_tfmid) + "&token=" + api_token,
					headers = {
						"User-Agent": "Mozilla/5.0"
					}
				)
			).read()
		)
		
		if result["success"]:
			if not result["internal_error"]:
				self.game_version = result["version"]
				self.game_connection_key = result["connection_key"]
				self.game_auth_key = result["auth_key"]
				self.game_packet_keys = result["packet_keys"]
				self.game_identification_keys = result["identification_keys"]
				self.game_msg_keys = result["msg_keys"]
				SetPacketKeys([self.game_packet_keys, self.game_identification_keys, self.game_msg_keys])
			
			else:
				if result["internal_error_step"] == 2:
					mightBeMaintenance = ": The game might be in maintenance mode."
				else:
					mightBeMaintenance = ""
				raise ApiEndpointException("There was an internal API Endpoint error. (" + str(result["internal_error_step"]) + ")" + mightBeMaintenance)
		else:
			raise ApiEndpointException("Can't get the keys. Error info: " + result["error"])
	
	def start(self, api_tfmid, api_token):
		"""Starts the bot.
		Gets the encryption keys, starts a connection, sends the handshake packet and starts the main loop."""
		
		self.get_keys(api_tfmid, api_token)
		
		self.main.connect("164.132.202.12", 5555)
		
		packet = ByteArray().writeShort(self.game_version).writeUTF(self.game_connection_key)
		packet.writeUTF("Desktop").writeUTF("-").writeLong(8125).writeUTF("")
		packet.writeUTF("86bd7a7ce36bec7aad43d51cb47e30594716d972320ef4322b7d88a85904f0ed")
		packet.writeUTF("A=t&SA=t&SV=t&EV=t&MP3=t&AE=t&VE=t&ACC=t&PR=t&SP=f&SB=f&DEB=f&V=LNX 29,0,0,140&M=Adobe Linux&R=1920x1080&COL=color&AR=1.0&OS=Linux&ARCH=x86&L=en&IME=t&PR32=t&PR64=t&LS=en-US&PT=Desktop&AVD=f&LFD=f&WD=f&TLS=t&ML=5.1&DP=72")
		packet.writeLong(0).writeLong(25175).writeUTF("")
		
		self.main.send([28, 1], packet)
		
		self.SmartThread.runParallelTask(target = self.SmartThread.tasksQueueLoop, args = (self.main,))
		self.loop_run()